==> cdns-rs-0.1.1.crate: cdns-rs-0.1.1/src/sync/caches.rs <==
/*-
* cdns-rs - a simple sync/async DNS query library
* Copyright (C) 2020  Aleksandr Morozov, RELKOM s.r.o
* Copyright (C) 2021  Aleksandr Morozov
* 
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 3 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
* 
* You should have received a copy of the GNU Lesser General Public License
* along with this program; if not, write to the Free Software Foundation,
* Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

/// This file contains configuration caches.

use std::collections::HashSet;
use std::sync::Arc;
use std::time::SystemTime;
use std::fs;


use crate::{write_error, internal_error, internal_error_map, error::*};

use super::cfg_parsers::*;
use super::mutex::*;
use super::common::{RESOLV_CFG_PATH, HOST_CFG_PATH};

lazy_static!{
    pub static ref CACHE: CachesController = CachesController::new();
}

#[derive(Clone, Debug)]
struct CacheTime<'cache>
{
    path: &'cache str,
    last_modif: SystemTime,
}

impl<'cache> CacheTime<'cache>
{
    fn new(path: &'cache str) -> CDnsResult<Self>
    {
        return Ok(Self { path: path, last_modif: Self::get_last_modified(path)? });
    }

    fn fake(path: &'cache str) -> Self
    {
        return Self { path: path, last_modif: SystemTime::now() };
    }

    fn get_last_modified(path: &'cache str) -> CDnsResult<SystemTime>
    {
        let metadata = fs::metadata(path).map_err(|e| 
            internal_error_map!(CDnsErrorType::InternalError, "fs::metadata::modified() not supported on this platform: '{}'", e)
        )?;

        let last_modif = 
            metadata.modified().map_err(|e| 
                internal_error_map!(CDnsErrorType::InternalError, "fs::metadata::modified() not supported on this platform: '{}'", e)
            )?;

        return Ok(last_modif);
    }

    fn check_modified(&mut self) -> CDnsResult<bool>
    {
        // in case if get_last_modified() will return Err, return what was cached prevously
        let last_modif = 
            match Self::get_last_modified(self.path)
            {
                Ok(t) => t,
                Err(e) => internal_error!(CDnsErrorType::InternalError, "{}", e),
            };

        return Ok(self.last_modif != last_modif);
        /*{
            // reload
            match ResolveConfEntry::parse_resolve_cfg()
            {
                Ok(r) => 
                    self.cache = Arc::new(r),
                Err(e) =>
                    write_error!("{}", e),
            }
             
        }
        
        return;*/
    }
}

#[derive(Clone, Debug)]
pub struct ConfigCacheResolve
{
    last_modif: CacheTime<'static>,
    cache: Arc<Vec<ResolveConfEntry>>,
}

impl Default for ConfigCacheResolve
{
    fn default() -> Self 
    {
        return Self {last_modif: CacheTime::fake(RESOLV_CFG_PATH), cache: Arc::new(Vec::new()) };
    }
}

impl ConfigCacheResolve
{
    fn new() -> CDnsResult<Self>
    {
        return Ok(
            Self 
            {
                last_modif: CacheTime::new(RESOLV_CFG_PATH)?, 
                cache: Self::read_config()?
            }
        );
    }

    fn read_config() -> CDnsResult<Arc<Vec<ResolveConfEntry>>>
    {
        return Ok(Arc::new(ResolveConfEntry::parse_resolve_cfg()?));
    }

    fn check_modified(&mut self) -> CDnsResult<()>
    {
        // in case if get_last_modified() will return Err, save what was cached prevously

        if self.last_modif.check_modified()? == true
        {
            // reload
            match ResolveConfEntry::parse_resolve_cfg()
            {
                Ok(r) => 
                    self.cache = Arc::new(r),
                Err(e) =>
                    return Err(e),
            }
             
        }
        
        return Ok(());
    }

    fn clone_cache(&self) -> Arc<Vec<ResolveConfEntry>>
    {
        return self.cache.clone();
    }
}

#[derive(Clone, Debug)]
pub struct ConfigCacheHosts
{
    last_modif: CacheTime<'static>,
    cache: Arc<HashSet<HostnameEntry>>,
}

impl Default for ConfigCacheHosts
{
    fn default() -> Self 
    {
        return Self {last_modif: CacheTime::fake(HOST_CFG_PATH), cache: Arc::new(HashSet::new()) };
    }
}

impl ConfigCacheHosts
{
    fn new() -> CDnsResult<Self>
    {
        return Ok(
            Self {last_modif: CacheTime::new(HOST_CFG_PATH)?, cache: Self::read_config()? }
        );
    }

    fn read_config() -> CDnsResult<Arc<HashSet<HostnameEntry>>>
    {
        return Ok(
            Arc::new(
            HostnameEntry::parse_host_file(true)?.into_iter().map(|f| f).collect()
            )
        );
    }

    fn check_modified(&mut self) -> CDnsResult<()>
    {
        // in case if get_last_modified() will return Err, save what was cached prevously

        if self.last_modif.check_modified()? == true
        {
            // reload
            match HostnameEntry::parse_host_file(true)
            {
                Ok(r) =>
                    self.cache = Arc::new(r.into_iter().map(|f| f).collect()),
                Err(e) =>
                    return Err(e),
            }
        }
        

        return Ok(());
    }

    fn clone_cache(&self) -> Arc<HashSet<HostnameEntry>>
    {
        return self.cache.clone();
    }
}

pub struct CachesController
{
    resolv_cache: Mutex<ConfigCacheResolve>,
    host_cache: Mutex<ConfigCacheHosts>
}

unsafe impl Sync for CachesController{}
unsafe impl Send for CachesController{}

impl CachesController
{
    fn new() -> Self
    {
        let resolve_cache = ConfigCacheResolve::new();

        let host_cache = ConfigCacheHosts::new();

        if resolve_cache.is_err() == true
        {
            write_error!("{}", resolve_cache.as_ref().err().unwrap());
        }
        
        if host_cache.is_err() == true
        {
            write_error!("{}", host_cache.as_ref().err().unwrap());
        }

        return CachesController
            { 
                resolv_cache: Mutex::new(resolve_cache.unwrap_or(ConfigCacheResolve::default())),
                host_cache: Mutex::new(host_cache.unwrap_or(ConfigCacheHosts::default())),
            };
    }

    pub 
    fn is_resolve_cached(&self) -> bool
    {
        return !self.resolv_cache.lock().unwrap().cache.is_empty();
    }

    pub 
    fn is_host_cached(&self) -> bool
    {
        return !self.host_cache.lock().unwrap().cache.is_empty();
    }

    fn try_cache_resolve(&self) -> CDnsResult<Arc<Vec<ResolveConfEntry>>>
    {
        let mut mutx_resolv = self.resolv_cache.lock().unwrap();
        mutx_resolv.check_modified()?;

        return Ok(mutx_resolv.clone_cache());
    }

    pub 
    fn clone_resolve_list(&self) -> CDnsResult<Arc<Vec<ResolveConfEntry>>>
    {
        // clone from cache and/or reload cache
        match self.try_cache_resolve()
        {
            Ok(r) => return Ok(r),
            Err(e) => 
            {
                write_error!("{}", e);
                // try to read directly
                return ConfigCacheResolve::read_config();
            }
        }
    }

    fn try_cache_host(&self) -> CDnsResult<Arc<HashSet<HostnameEntry>>>
    {
        let mut mutx_host = self.host_cache.lock().unwrap();
        mutx_host.check_modified()?;

        return Ok(mutx_host.clone_cache());
    }

    pub 
    fn clone_host_list(&self) -> CDnsResult<Arc<HashSet<HostnameEntry>>>
    {
        match self.try_cache_host()
        {
            Ok(r) => return Ok(r),
            Err(e) =>
            {
                write_error!("{}", e);
                // try to read directly
                return ConfigCacheHosts::read_config();
            }
        }
    }
}


#[test]
fn test_init()
{
    let res = CACHE.clone_resolve_list();
    assert_eq!(res.is_ok(), true, "{}", res.err().unwrap());

    let res = res.unwrap();

    println!("{:?}", res);

    let res = CACHE.clone_host_list();
    assert_eq!(res.is_ok(), true, "{}", res.err().unwrap());

    let res = res.unwrap();

    println!("{:?}", res);
}

==> cdns-rs-0.2.0.crate: cdns-rs-0.2.0/src/sync/caches.rs <==
/*-
* cdns-rs - a simple sync/async DNS query library
* Copyright (C) 2020  Aleksandr Morozov, RELKOM s.r.o
* Copyright (C) 2021-2022  Aleksandr Morozov
* 
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 3 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
* 
* You should have received a copy of the GNU Lesser General Public License
* along with this program; if not, write to the Free Software Foundation,
* Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

/// This file contains configuration caches.

use std::path::Path;
use std::sync::Arc;
use std::time::SystemTime;
use std::fs;


use crate::cfg_host_parser::HostConfig;
use crate::cfg_resolv_parser::ResolveConfig;
use crate::{write_error, internal_error, internal_error_map, error::*};

use super::cfg_parsers::*;
use super::mutex::*;

lazy_static!{
    pub static ref CACHE: CachesController = CachesController::new();
}


pub trait CacheOperations
{
    fn is_reload_allowed(&self) -> bool;
}

#[derive(Clone, Debug)]
struct CacheTime<'cache>
{
    path: &'cache Path,
    last_modif: SystemTime,
}

impl<'cache> CacheTime<'cache>
{
    fn new(path: &'cache Path) -> CDnsResult<Self>
    {
        return Ok(Self { path: path, last_modif: Self::get_last_modified(path)? });
    }

    fn fake(path: &'cache Path) -> Self
    {
        return Self { path: path, last_modif: SystemTime::now() };
    }

    fn get_last_modified(path: &'cache Path) -> CDnsResult<SystemTime>
    {
        let metadata = fs::metadata(path).map_err(|e| 
            internal_error_map!(CDnsErrorType::InternalError, "fs::metadata::modified() not supported on this platform: '{}'", e)
        )?;

        let last_modif = 
            metadata.modified().map_err(|e| 
                internal_error_map!(CDnsErrorType::InternalError, "fs::metadata::modified() not supported on this platform: '{}'", e)
            )?;

        return Ok(last_modif);
    }

    fn check_modified(&mut self) -> CDnsResult<bool>
    {
        // in case if get_last_modified() will return Err, return what was cached prevously
        let last_modif = 
            match Self::get_last_modified(self.path)
            {
                Ok(t) => t,
                Err(e) => internal_error!(CDnsErrorType::InternalError, "{}", e),
            };

        return Ok(self.last_modif != last_modif);
        /*{
            // reload
            match ResolveConfEntry::parse_resolve_cfg()
            {
                Ok(r) => 
                    self.cache = Arc::new(r),
                Err(e) =>
                    write_error!("{}", e),
            }
             
        }
        
        return;*/
    }
}

#[derive(Clone, Debug)]
pub struct CacheInstance<T: Default + ConfigParser<T> + CacheOperations>//, P: ConfigParser<T>>
{
    last_modif: CacheTime<'static>,
    cache: Arc<T>,
   // f: PhantomData<P>,
}

impl<T: Default + ConfigParser<T> + CacheOperations> Default for CacheInstance<T>
{
    fn default() -> Self 
    {
        return 
            Self 
            {
                last_modif: CacheTime::fake(T::get_file_path()), 
                cache: Arc::new(T::default()),
              //  f: PhantomData,
            };
    }
}

impl<T: Default + ConfigParser<T> + CacheOperations> CacheInstance<T>
{
    fn new() -> CDnsResult<Self>
    {
        return Ok(
            Self 
            {
                last_modif: CacheTime::new(T::get_file_path())?, 
                cache: Arc::new(T::parse_config()?),
                //f: PhantomData,
            }
        );
    }

    fn read_config() -> CDnsResult<Arc<T>>
    {
        return Ok(Arc::new(T::parse_config()?));
    }

    fn check_modified(&mut self) -> CDnsResult<()>
    {
        // in case if get_last_modified() will return Err, save what was cached prevously
        if self.cache.is_reload_allowed() == false
        {
            // just return OK
            return Ok(());
        }

        if self.last_modif.check_modified()? == true
        {
            // reload
            self.cache = Self::read_config()?;
        }
        
        return Ok(());
    }

    fn clone_cache(&self) -> Arc<T>
    {
        return self.cache.clone();
    }
}

pub struct CachesController
{
    resolv_cache: Mutex<CacheInstance<ResolveConfig>>,
    host_cache: Mutex<CacheInstance<HostConfig>>
}

unsafe impl Sync for CachesController{}
unsafe impl Send for CachesController{}

impl CachesController
{
    /// Spawns instance and reads the configuration.
    fn new() -> Self
    {
        let resolve_cache_res = CacheInstance::<ResolveConfig>::new();

        let host_cache_res = CacheInstance::<HostConfig>::new();

        let resolve_cache = 
            match resolve_cache_res
            {
                Ok(r) => r,
                Err(e) =>
                {
                    write_error!("{}", e);

                    CacheInstance::default()
                }
            };
        
        let host_cache = 
            match host_cache_res
            {
                Ok(r) => r,
                Err(e) =>
                {
                    write_error!("{}", e);

                    CacheInstance::default()
                }
            };
    

        return 
            CachesController
            { 
                resolv_cache: Mutex::new(resolve_cache),
                host_cache: Mutex::new(host_cache),
            };
    }

    pub 
    fn is_resolve_cached(&self) -> bool
    {
        return !self.resolv_cache.lock().unwrap().cache.is_default();
    }

    pub 
    fn is_host_cached(&self) -> bool
    {
        return !self.host_cache.lock().unwrap().cache.is_default();
    }

    fn try_cache_resolve(&self) -> CDnsResult<Arc<ResolveConfig>>
    {
        let mut mutx_resolv = self.resolv_cache.lock().unwrap();
        mutx_resolv.check_modified()?;

        return Ok(mutx_resolv.clone_cache());
    }

    /// Clones the [ResolveConfig] [Arc] (atomic reference counter) reference.
    pub 
    fn clone_resolve_list(&self) -> CDnsResult<Arc<ResolveConfig>>
    {
        // clone from cache and/or reload cache
        match self.try_cache_resolve()
        {
            Ok(r) => return Ok(r),
            Err(e) => 
            {
                write_error!("{}", e);

                // try to read directly, this ignores all reload restrictions
                return CacheInstance::<ResolveConfig>::read_config();
            }
        }
    }

    fn try_cache_host(&self) -> CDnsResult<Arc<HostConfig>>
    {
        let mut mutx_host = self.host_cache.lock().unwrap();
        mutx_host.check_modified()?;

        return Ok(mutx_host.clone_cache());
    }

    /// Clones the [HostConfig] [Arc] (atomic reference counter) reference.
    pub 
    fn clone_host_list(&self) -> CDnsResult<Arc<HostConfig>>
    {
        match self.try_cache_host()
        {
            Ok(r) => return Ok(r),
            Err(e) =>
            {
                write_error!("{}", e);

                // try to read directly, this ignores all reload restrictions
                return CacheInstance::<HostConfig>::read_config();
            }
        }
    }
}


#[test]
fn test_init()
{
    let res = CACHE.clone_resolve_list();
    assert_eq!(res.is_ok(), true, "{}", res.err().unwrap());

    let res = res.unwrap();

    println!("{:?}", res);

    let res = CACHE.clone_host_list();
    assert_eq!(res.is_ok(), true, "{}", res.err().unwrap());

    let res = res.unwrap();

    println!("{:?}", res);
}
